@@ -9,12 +9,11 @@ from typing import List
 
 from .config import CRFConfig
 from .feature_extraction import (
-    sent2features,
-    sent2labels,
-    FEATURE_LAYER_WORD,
-    FEATURE_LAYER_LEMMA,
+    FeatureExtractor,
     FEATURE_LAYER_POS,
     FEATURE_LAYER_GRAMMAR,
+    FEATURE_LAYER_WORD,
+    KNOWN_LAYERS
 )
 from .tagging_report import TaggingReport
 from . import exceptions
@@ -33,7 +32,6 @@ def check_model_loaded(func):
 
 
 class CRFExtractor:
-    MLP_FIELDS = [FEATURE_LAYER_WORD, FEATURE_LAYER_LEMMA, FEATURE_LAYER_POS, FEATURE_LAYER_GRAMMAR]
 
     def __init__(self,
         description: str = "My CRF Extractor",
@@ -42,7 +40,8 @@ class CRFExtractor:
     ):
         self.description = description
         self.config = config
-        self.embedding = embedding
+        self.feature_extractor = FeatureExtractor(config, embedding=embedding)
+
         self.model = None
         self.best_c_values = None
 
@@ -51,53 +50,12 @@ class CRFExtractor:
         return self.description
 
 
-    def _validate_and_extract_layers(self, mlp_document: dict, field_name: str):
-        """
-        Validates the MLP document structure to have proper fields.
-        """
-        field_path = field_name.split(".")
-        # parse field path
-        for field_path_component in field_path:
-            if field_path_component not in mlp_document:
-                raise exceptions.InvalidInputError(
-                    f"Invalid field_name param for the document. Field component {field_path_component} not found in document!"
-                )
-            mlp_subdoc = mlp_document[field_path_component]
-        # check the resulting subdocument structure
-        if not isinstance(mlp_subdoc, dict):
-            raise exceptions.InvalidInputError("Document is not a dict!")
-        # empty dict for layers
-        layers = {}
-        fields_to_check = set(self.config.feature_layers + self.config.context_feature_layers)
-        # check if mlp fields are present
-        for field in fields_to_check:
-            if field not in CRFExtractor.MLP_FIELDS:
-                raise exceptions.InvalidInputError(f"Field '{field}' is an unknown MLP field!")
-            if field not in mlp_subdoc:
-                raise exceptions.InvalidInputError(f"Field '{field}' not present in the document!")
-            # add layer if not present
-            if field not in layers:
-                layers[field] = []
-            # split field into sents
-            # use "LBR" as sentence break marker for POS tag layer
-            if field in (FEATURE_LAYER_POS, FEATURE_LAYER_GRAMMAR):
-                sentences = mlp_subdoc[field].split(" LBR ")
-            else:
-                sentences = mlp_subdoc[field].split(" \n ")
-            # add sent to layer
-            for sentence in sentences:
-                sentence_tokens = sentence.split(" ")
-                layers[field].append(sentence_tokens)
-        return layers
-
-
     def _parse_mlp_document(self, mlp_document: dict, add_labels: bool = True, mlp_field: str = "text_mlp"):
         """
         Parses MLP output document. Extracts tokens, lemmas, and POS tags.
         Adds labels from texta_facts.
         """
-        # TODO: add morph/syntactic info if possible (need to improve MLP for that)
-        layers = self._validate_and_extract_layers(mlp_document, mlp_field)
+        layers = self.feature_extractor.validate_and_extract_layers(mlp_document, mlp_field)
         # add labels from texta_facts
         if add_labels:
             labels = []
@@ -139,17 +97,13 @@ class CRFExtractor:
         for mlp_document in mlp_documents:
             for sentence in self._parse_mlp_document(mlp_document, mlp_field=mlp_field):
                 token_documents.append(sentence)
-
         # split dataset
         train_documents, test_documents = train_test_split(token_documents, test_size=self.config.test_size)
         # featurize sequences
-        X_train = [sent2features(s, self.config, self.embedding) for s in train_documents]
-        y_train = [sent2labels(s) for s in train_documents]
-        X_test = [sent2features(s, self.config, self.embedding) for s in test_documents]
-        y_test = [sent2labels(s) for s in test_documents]
-
-
-        
+        X_train = [self.feature_extractor.sent2features(s) for s in train_documents]
+        y_train = [self.feature_extractor.sent2labels(s) for s in train_documents]
+        X_test = [self.feature_extractor.sent2features(s) for s in test_documents]
+        y_test = [self.feature_extractor.sent2labels(s) for s in test_documents]
         # Test different C-values to get optimal model.
         best_f1 = 0
         best_report = None
@@ -178,8 +132,7 @@ class CRFExtractor:
                 # evaluate model
                 y_pred = [tagger.tag(xseq) for xseq in X_test]
                 report = TaggingReport(y_test, y_pred, self.config.labels)
-
-                print(report.f1_score)
+                #print(report.f1_score)
                 # check if model better than our best
                 if report.f1_score > best_f1:
                     # found a new best model!
@@ -211,7 +164,7 @@ class CRFExtractor:
         for i, seq_to_predict in seqs_to_predict:
             # make input edible for feature extractor
             seq_to_predict = [(a, None) for a in seq_to_predict]
-            features_to_predict = sent2features(seq_to_predict, self.config, self.embedding)
+            features_to_predict = self.feature_extractor.sent2features(seq_to_predict)
             result = self.model.tag(features_to_predict)
             # generate text tokens for final output
             for tag in self._process_tag_output(result, sentences, i, field_name):
@@ -227,12 +180,14 @@ class CRFExtractor:
 
     @check_model_loaded
     def get_features(self, n=20):
+        """
+        Retrieves positive & negative features for the model.
+        """
         info = self.model.info()
         all_feats = Counter(info.state_features).most_common()
-
+        # filter features
         positive_features = [{"feature": f[0], "coefficient": c, "label": f[1]} for f, c in Counter(info.state_features).most_common(n)]
         negative_features = [{"feature": f[0], "coefficient": c, "label": f[1]} for f, c in all_feats[-n:]]
-
         return {
             "features": {
                 "positive": positive_features,
